use std::{
    collections::{HashMap, HashSet, VecDeque},
    fmt::Pointer,
    pin::pin,
    sync::Arc,
    task::{ready, Poll},
};

use futures::{
    stream::{SplitSink, SplitStream},
    Sink, SinkExt, Stream, StreamExt,
};
use jabber::{
    connection::Tls,
    jabber_stream::bound_stream::{BoundJabberReader, BoundJabberStream, BoundJabberWriter},
    JID,
};
use sqlx::{query, Pool, Sqlite, SqlitePool};
use stanza::{
    client::{
        iq::{Iq, IqType, Query},
        Stanza,
    },
    roster,
};
use tokio::{
    io::AsyncRead,
    select,
    sync::{
        mpsc::{self, Receiver, Sender},
        Mutex,
    },
    task::{JoinHandle, JoinSet},
};
use tokio_stream::wrappers::ReceiverStream;
use tokio_util::sync::{PollSendError, PollSender};

pub struct Luz {
    receiver: ReceiverStream<UpdateMessage>,
    sender: PollSender<CommandMessage>,
    // TODO: oneshot
    pending_iqs: Arc<Mutex<HashMap<String, mpsc::Sender<Iq>>>>,
    db: SqlitePool,
    tasks: JoinSet<()>,
    jid: JID,
}

impl Luz {
    pub async fn new() {}

    pub async fn supervisor(
        mut read_sender: Sender<UpdateMessage>,
        mut write_receiver: Receiver<CommandMessage>,
        pending_iqs: Arc<Mutex<HashMap<String, mpsc::Sender<Iq>>>>,
        db: SqlitePool,
        jid: &JID,
    ) {
        let connection: Arc<
            Mutex<Option<(JoinHandle<()>, (JoinHandle<()>, Sender<CommandMessage>))>>,
        > = Arc::new(Mutex::new(None));
        let mut jid = jid.clone();
        let mut server = jid.domainpart.clone();
        let tasks: Arc<Mutex<JoinSet<()>>> = Arc::new(Mutex::new(JoinSet::new()));
        for command in write_receiver.recv().await {
            match command {
                CommandMessage::Connect(_) => todo!(),
                CommandMessage::Disconnect => todo!(),
                CommandMessage::Kill => todo!(),
                CommandMessage::GetRoster => todo!(),
                CommandMessage::SendMessage(jid, _) => todo!(),
            }
            tasks.lock().await.spawn(handle_command(connection.clone()));
        }
    }
}

pub async fn handle_command(
    connection: Arc<Mutex<Option<(JoinHandle<()>, (JoinHandle<()>, Sender<CommandMessage>))>>>,
) {
    match command {
        CommandMessage::Connect(password) => {
            let streams = jabber::connect_and_login(&mut jid, password, &mut server)
                .await
                .unwrap();
            let (read, write) = streams.split();
            let (sender, receiver) = mpsc::channel(20);
            let reads = tokio::spawn(jabber_reads(read, read_sender.clone(), db.clone()));
            let writes = tokio::spawn(jabber_writes(write, receiver, db.clone()));
            // properly handle previous connection
            *connection.lock().await = Some((reads, (writes, sender)));
        }
        CommandMessage::Disconnect => {
            if let Some((reads, (writes, write_channel))) = connection.lock().await.take() {
                write_channel.send(CommandMessage::Disconnect).await;
                writes.await;
                reads.await;
                jid = jid.as_bare();
            }
        }
        CommandMessage::GetRoster => match *connection.lock().await {
            Some(_) => todo!(),
            None => todo!(),
        },
        CommandMessage::SendMessage(jid, _) => todo!(),
    }
}

pub async fn jabber_writes(
    mut jabber_writer: BoundJabberWriter<Tls>,
    mut write_receiver: Receiver<CommandMessage>,
    /*error_sender: Sender<UpdateMessage>,*/ db: SqlitePool,
) {
}

pub async fn jabber_reads<S>(
    mut read: BoundJabberReader<S>,
    read_sender: Sender<UpdateMessage>,
    db: SqlitePool,
) where
    S: AsyncRead + Unpin + Sync,
{
    while let Ok(stanza) = read.read::<Stanza>().await {
        let upd = match stanza {
            Stanza::Message(message) => todo!(),
            Stanza::Presence(presence) => todo!(),
            Stanza::Iq(iq) => match iq.r#type {
                IqType::Error => todo!(),
                IqType::Get => todo!(),
                IqType::Result => match iq.query {
                    Some(q) => match q {
                        Query::Bind(bind) => todo!(),
                        Query::Ping(ping) => todo!(),
                        Query::Roster(query) => UpdateMessage::Roster(query.items),
                        Query::Unsupported => todo!(),
                    },
                    None => todo!(),
                },
                IqType::Set => todo!(),
            },
            Stanza::Error(error) => todo!(),
            Stanza::OtherContent(content) => todo!(),
        };
        read_sender.send(upd).await.expect("read_sender");
    }
    todo!()
}

pub enum JabberWriteCommand {}

pub async fn funnel(
    mut jabber_reads: Receiver<UpdateMessage>,
    mut db_reads: Receiver<UpdateMessage>,
    out: Sender<UpdateMessage>,
) {
    select! {
        j = jabber_reads.recv() => {
            out.send(j.unwrap()).await;
        },
        d = db_reads.recv() => {
            out.send(d.unwrap()).await;
        }
    }
}

pub async fn write_thread(mut write: BoundJabberWriter<S>, write_recv: Receiver<CommandMessage>) {
    while let Some(cmd) = write_recv.recv().await {
        match cmd {
            CommandMessage::GetRoster => todo!(),
            CommandMessage::SendMessage(jid, _) => todo!(),
            CommandMessage::Connect => continue,
            CommandMessage::Disconnect => write_thread_disconnected(write_recv),
        }
    }
    todo!()
}

impl Client {
    pub fn send_command(self: Arc<Self>, cmd: CommandMessage) {
        match cmd {
            CommandMessage::Connect => todo!(),
            CommandMessage::Disconnect => todo!(),
            CommandMessage::GetRoster => self.get_roster(),
            CommandMessage::SendMessage(jid, _) => todo!(),
        }
    }
    async fn begin_query_roster(self: Arc<Self>) -> Result<()> {
        if self.connected {
            self.db.start_get_roster().await;
            self.write_chan.send(CommandMessage::GetRoster).await;
        } else {
            let roster = self.db.get_roster();
            self.db_chan.send(roster).await;
        }
    }
}

pub struct ReadStreamThread {
    // held by both read thread and write thread
    pending_iqs: Arc<Mutex<HashMap<String, mpsc::Sender<Iq>>>>,
    tasks: JoinSet<()>,
}

pub struct WriteStreamThread {
    pending_iqs: Arc<Mutex<HashMap<String, mpsc::Sender<Iq>>>>,
    tasks: JoinSet<()>,
}

impl Luz {
    // pub async fn new(jid: &str, dburl: &str) -> Result<Self, Error> {
    //     let client = Client::new(jid, dburl).await?;
    //     let luz = Arc::new(Mutex::new(client));
    //     client
    //         .tasks
    //         .spawn(Client::read_thread(luz.clone(), read_sender));
    //     client.tasks.spawn(Client::write_thread(write_receiver));
    //     let luz = luz.into_inner();
    //     Ok(Luz(luz))
    // }
}

// instead of having a channel to send stanzas down, have a channel to send channels that you can send stanzas down, so if one channel fails (through disconnection, etc), you can move on to the next. could also have an enum to let them sit until reconnection.
impl Client {
    async fn new(jid: &str, dburl: &str) -> Result<Self, Error> {
        let (read_sender, read_receiver) = mpsc::channel::<UpdateMessage>(20);
        let (write_sender, write_receiver) = mpsc::channel::<CommandMessage>(20);
        let jid: JID = jid.try_into()?;
        let server = jid.domainpart.clone();
        let db = SqlitePool::connect(dburl).await?;
        // let bound_streams = jabber::connect_and_login(&mut jid, password, &mut server).await?;
        // let (read, write) = bound_streams.split();
        // let streams = SplitStreams { read: Mutex::new(read), write: Mutex::new(write) };
        let mut tasks = JoinSet::new();
        let client = Self {
            receiver: ReceiverStream::new(read_receiver),
            sender: PollSender::new(write_sender),
            pending_iqs: HashMap::new(),
            db,
            jid,
            server,
            // TODO: check for drop abort handles
            tasks,
        };
        // store handles in client
        // send connection streams down channel created here
        Ok(client)
    }

    pub async fn read_thread(client: Arc<Mutex<Self>>, sender: mpsc::Sender<UpdateMessage>) {}

    pub async fn write_thread(receiver: mpsc::Receiver<CommandMessage>) {
        for message in receiver.recv().await {}
    }

    pub async fn connect(&mut self, password: &str) -> Result<(), Error> {
        todo!()
    }
}

pub enum Error {
    PollSend(PollSendError<CommandMessage>),
    Jabber(jabber::Error),
    SQL(sqlx::Error),
    JID(jid::ParseError),
}

impl From<jid::ParseError> for Error {
    fn from(e: jid::ParseError) -> Self {
        Self::JID(e)
    }
}

impl From<sqlx::Error> for Error {
    fn from(e: sqlx::Error) -> Self {
        Self::SQL(e)
    }
}

impl From<jabber::Error> for Error {
    fn from(e: jabber::Error) -> Self {
        Self::Jabber(e)
    }
}

impl Stream for Client {
    type Item = UpdateMessage;

    fn poll_next(
        self: std::pin::Pin<&mut Self>,
        cx: &mut std::task::Context<'_>,
    ) -> std::task::Poll<Option<Self::Item>> {
        pin!(self).receiver.poll_next_unpin(cx)
    }
}

impl From<PollSendError<CommandMessage>> for Error {
    fn from(e: PollSendError<CommandMessage>) -> Self {
        Self::PollSend(e)
    }
}

pub enum CommandMessage {
    Connect(String),
    Disconnect,
    /// gets the roster. if offline, retreives cached version from database. should be stored in application memory.
    GetRoster,
    SendMessage(JID, String),
}

pub enum UpdateMessage {
    Roster(Vec<roster::Item>),
}

impl Client {
    pub async fn process_stanza(
        &mut self,
        stanza: Result<Stanza, jabber::Error>,
        sender: mpsc::Sender<UpdateMessage>,
    ) {
        match stanza {
            Ok(stanza) => todo!(),
            Err(e) => self.process_error(e),
        }
    }

    pub async fn iq(
        &mut self,
        to: Option<JID>,
        r#type: IqType,
        query: Option<Query>,
    ) -> Result<IqResponse, Error> {
        self.client
            .send(Stanza::Iq(Iq {
                from: Some(self.client.jid()),
                // TODO: generate id
                id: "test".to_string(),
                to,
                r#type,
                // TODO: lang
                lang: None,
                query,
                errors: Vec::new(),
            }))
            .await?;
        Ok(todo!())
    }

    pub async fn iq_process(&mut self, iq: Iq) {}
}
